"""
エクセルで計算した梁特殊荷重をSS7に取りこみます。
詳細は README.pdf を参照してください。

Copyright (C) 2024 UNION SYSTEM Inc.

"""

import numpy as np
import os
import csv
import openpyxl
from openpyxl.worksheet.worksheet import Worksheet
from typing import Tuple
from Python import Ss7Python as Cmd

# 定数の定義
LOAD_MAX = 5  # 荷重の最大数
CSV_PATH = os.path.join(
    os.path.dirname(os.getcwd()), "docs", "Imput.csv"   # CSVのパス
)
EXCEL_PATH = os.path.join(
    os.path.dirname(os.getcwd()), "docs", "特殊荷重計算表.xlsx"     # 特殊荷重表のエクセルデータのパス
)
SS7_PATH = os.path.join(
    os.path.dirname(os.getcwd()), "docs", "Sample1.ikn"     # 元データのss7パス
)
SS7_ADDLOAD_PATH = os.path.join(
    os.path.dirname(os.getcwd()), "docs", "Sample1_特殊荷重追加.ikn"    # 新規作成するss7パス
)
COLUMN = ["B", "M", "O", "P", "R", "S"]     # エクセルの列


load_name = np.full((LOAD_MAX,), "", dtype="U24")   # 荷重名称
load_type = np.full((LOAD_MAX,), "", dtype="U10")   # タイプ
P = np.full((LOAD_MAX, 6), "", dtype="U10")         # P1,P2,P3,P4,P5,P6
LL_TL = np.full((LOAD_MAX,), "", dtype="U4")        # LL/TL
earth_ramen = np.full((LOAD_MAX,), "", dtype="U4")  # 地/ラ
export_judge = np.full((LOAD_MAX,), "", dtype="U10")        # データ転送判定
overwrite_judge = np.full((LOAD_MAX,), "", dtype="U10")     # 上書き判定


def read_excel_data() -> None:
    """エクセルからデータを取得する"""
    # Excelファイルを開く
    workbook = openpyxl.load_workbook(EXCEL_PATH, data_only=True)
    # シートを選択する
    sheet = workbook["特殊荷重"]
    # 荷重入力の回数、処理を繰り返す
    for load_no in range(LOAD_MAX):
        row = [5 + i + load_no * 6 for i in range(6)]
        # 行列からそれぞれの値が入力してあるセル名を生成
        load_name_cell = COLUMN[0] + str(row[0])    # 荷重名称が入力してある箇所のセル
        load_type_cell = COLUMN[0] + str(row[1])    # タイプが入力してある箇所のセル
        P_cell = [COLUMN[1] + str(r) for r in row]  # P1～P6が入力してある箇所のセル
        LL_TL_cell = COLUMN[2] + str(row[0])        # LL/TLが入力してある箇所のセル
        earth_ramen_cell = COLUMN[3] + str(row[0])  # 地/ラが入力してある箇所のセル
        export_judge_cell = COLUMN[4] + str(row[0])     # データ転送判定が入力してある箇所のセル
        overwrite_judge_cell = COLUMN[5] + str(row[0])  # 上書き判定が入力してある箇所のセル

        # セルに入力してある値を取得
        load_name[load_no] = get_cell_value(sheet, load_name_cell)
        load_type[load_no] = get_load_type(sheet, load_type_cell)
        P[load_no] = get_P_values(sheet, P_cell)
        LL_TL[load_no] = get_cell_value(sheet, LL_TL_cell)
        earth_ramen[load_no] = get_cell_value(sheet, earth_ramen_cell)
        export_judge[load_no] = get_cell_value(sheet, export_judge_cell)
        overwrite_judge[load_no] = get_cell_value(sheet, overwrite_judge_cell)


def get_cell_value(sheet: Worksheet, cell: str) -> str:
    """
    セルの値を取得する

    Args:
        sheet (Worksheet): シート
        cell (str): セルの位置

    Returns:
        str: セルの値
    """
    return sheet[cell].value if sheet[cell].value else ""


def get_load_type(sheet: Worksheet, cell: str) -> str:
    """
    荷重タイプを取得する

    Args:
        sheet (Worksheet): シート
        cell (str): セルの位置

    Returns:
        str: 荷重タイプ
    """
    value = sheet[cell].value
    if value:
        # タイプ　入力CSVには1～14の番号を使用するため先頭の文字のみ抽出する
        load_type = value[0]
        # 2文字が:でない場合は荷重タイプ10～14が選択されていることになる
        if value[1] != ":":
            load_type += value[1]
        return load_type
    return ""


def get_P_values(sheet: Worksheet, cells: list[str]) -> np.ndarray:
    """
    P1～P6を取得する

    Args:
        sheet (Worksheet): シート
        cells (list[str]): セルの位置

    Returns:
        np.ndarray: P1～P6の値
    """
    return np.array(
        [sheet[cell].value if sheet[cell].value else "" for cell in cells]
    )


def find_matching_row() -> Tuple[int, int]:
    """
    入力CSVファイルから既存の特殊荷重入力を含む行を探す

    Returns:
        Tuple[int, int]: ターゲット文字列を含む行の行番号と特殊荷重の登録数
    """
    matching_row = 0
    load_count = -2
    with open(CSV_PATH) as file:
        reader = csv.reader(file)
        TARGET_STRING = "name=特殊荷重登録 梁"
        # 各行を検査して、target_stringが含まれていれば行番号を記録
        for row_num, row in enumerate(reader, start=1):
            if TARGET_STRING in "".join(row):
                matching_row = row_num
                # 空白行が来るまで処理を繰り返す
                for _ in range(1000):  # 1000は適当な上限
                    row_csv = next(reader)
                    # 空白が来た場合処理を抜ける
                    if not row_csv:
                        break
                    load_count += 1
                break
    return matching_row, load_count


def append_load_to_csv(matching_row: int, load_count: int) -> None:
    """
    入力CSVファイルにエクセルの荷重表を追加する

    Args:
        matching_row (int): ターゲット文字列を含む行の行番号
        load_count (int): 荷重の追加数
    """
    # 特殊荷重の入力がない場合
    if matching_row == 0:
        new_additions_loads()
    # 特殊荷重の入力がある場合
    else:
        #  同一名の特殊荷重の登録を見つける
        load_row, lno, lno_max = find_load_rows(matching_row)
        # 荷重を更新または追加する
        update_or_append_loads(
            matching_row, load_count, load_row, lno, lno_max
        )


def new_additions_loads() -> None:
    """
    入力CSVの最後に特殊荷重入力テーブルを追加する
    """
    with open(CSV_PATH, "a", newline="") as file:
        writer = csv.writer(file)
        # ヘッダー部分の情報をCSVファイルに書き込む
        new_rows = [
            ["name=特殊荷重登録 梁"],
            [
                "No.",
                "荷重名称",
                "タイプ",
                "P1",
                "P2",
                "P3",
                "P4",
                "P5",
                "P6",
                "CMoQoのみ",
                "LL/TL",
                "地/ラ",
            ],
            ["<data>"],
        ]
        writer.writerows(new_rows)
        # 荷重情報をCSVファイルに書き込む
        for load_no in range(LOAD_MAX):
            if export_judge[load_no] == "実行する":
                new_rows_load = [
                    load_no + 1,
                    load_name[load_no],
                    load_type[load_no],
                    P[load_no, 0],
                    P[load_no, 1],
                    P[load_no, 2],
                    P[load_no, 3],
                    P[load_no, 4],
                    P[load_no, 5],
                    "No",
                    LL_TL[load_no],
                    earth_ramen[load_no],
                    "<RE>",
                ]
                writer.writerow(new_rows_load)


def find_load_rows(matching_row: int) -> Tuple[np.ndarray, np.ndarray, int]:
    """
    同一名の特殊荷重の登録を見つける

    Args:
        matching_row (int): ターゲット文字列を含む行の行番号

    Returns:
        Tuple[np.ndarray, np.ndarray, int]: 荷重行、荷重番号、最大荷重番号
    """
    load_row = np.full((LOAD_MAX,), 0, dtype=int)   # 荷重名が一致している箇所の行数
    lno = np.full((LOAD_MAX,), 0, dtype=int)        # 荷重No
    lno_max = 0                                     # 最大の荷重No取得
    for load_no in range(LOAD_MAX):
        # データ転送が実行する以外が選択されている場合、処理を飛ばす
        if export_judge[load_no] != "実行する":
            continue
        with open(CSV_PATH) as file:
            reader = csv.reader(file)
            # 指定した行番号まで読み飛ばす
            for _ in range(matching_row + 2):
                next(reader)
            # 空白行が来るまで処理を繰り返す
            for count in range(1000):   # 1000は適当な上限
                row_csv = next(reader)
                # 荷重名が一致している時
                if row_csv and (load_name[load_no] in "".join(row_csv)):
                    load_row[load_no] = matching_row + 2 + count
                    lno[load_no] = int(row_csv[0])
                    break
                # 空白行が来た時
                elif not row_csv:
                    break
                # 既に登録されている荷重の中から最大の荷重No取得
                lno_max = max(lno_max, int(row_csv[0]))
    return load_row, lno, lno_max


def update_or_append_loads(
    matching_row: int,
    load_count: int,
    load_row: np.ndarray,
    lno: np.ndarray,
    lno_max: int,
) -> None:
    """
    荷重を更新または追加する

    Args:
        matching_row (int): ターゲット文字列を含む行の行番号
        load_count (int): 荷重の追加数
        load_row (np.ndarray): 荷重行
        lno (np.ndarray): 荷重番号
        lno_max (int): 最大荷重番号
    """
    for load_no in range(LOAD_MAX):
        # データ転送が実行する以外が選択されている場合、処理を飛ばす
        if export_judge[load_no] != "実行する":
            continue
        with open(CSV_PATH) as file:
            reader = csv.reader(file)
            SS7_data = list(reader)
            # 同一荷重名の登録があり、上書きするが選択されている場合上書きする
            if load_row[load_no] and overwrite_judge[load_no] == "上書きする":
                new_rows_load = [
                    lno[load_no],
                    load_name[load_no],
                    load_type[load_no],
                    P[load_no, 0],
                    P[load_no, 1],
                    P[load_no, 2],
                    P[load_no, 3],
                    P[load_no, 4],
                    P[load_no, 5],
                    "No",
                    LL_TL[load_no],
                    earth_ramen[load_no],
                    "<RE>",
                ]
                SS7_data[load_row[load_no]] = new_rows_load
                with open(CSV_PATH, "w", newline="") as file:
                    writer = csv.writer(file)
                    writer.writerows(SS7_data)
            # 同一荷重名の登録がない場合、特殊荷重リストの最終行に追加する
            elif load_row[load_no] == 0:
                lno_max += 1    # 荷重Noの重複が無いように＋1する
                new_rows_load = [
                    lno_max,
                    load_name[load_no],
                    load_type[load_no],
                    P[load_no, 0],
                    P[load_no, 1],
                    P[load_no, 2],
                    P[load_no, 3],
                    P[load_no, 4],
                    P[load_no, 5],
                    "No",
                    LL_TL[load_no],
                    earth_ramen[load_no],
                    "<RE>",
                ]
                SS7_data.insert(matching_row + 2 + load_count, new_rows_load)
                with open(CSV_PATH, "w", newline="") as file:
                    writer = csv.writer(file)
                    writer.writerows(SS7_data)
                load_count += 1     # 登録したので荷重数のカウントを+1する


def export_csv_from_ss7() -> None:
    """SS7データを開き入力CSVファイルを出力する"""
    Cmd.Init()
    Cmd.Start("1.1.1.19", 1)
    SS7_data = Cmd.Open(SS7_PATH, 1, 3)
    if SS7_data is not None:
        inp = SS7_data.GetInputData()
        inp.ExportInputCsv(CSV_PATH, 1, 2)
        err = Cmd.GetLastError()
        if err.IsOK():
            print("入力CSVファイルのエクスポートが成功しました")
        else:
            print(f"エラー: {err.GetErrorMessage()}")
        SS7_data.Close(2)
    else:
        print("データが開けませんでした")
    Cmd.End(1)


def create_ss7_from_csv() -> None:
    """入力CSVファイルからSS7ファイルを新規作成する"""
    Cmd.Init()
    Cmd.Start("1.1.1.19", 1)
    Cmd.CreateDataCsv(CSV_PATH, SS7_ADDLOAD_PATH, 2)
    err = Cmd.GetLastError()
    if err.IsOK():
        print("CSVデータ新規作成が成功しました")
    else:
        print("CSVデータ新規作成が失敗しました。")


def main() -> None:
    """メイン処理"""
    # エクセルからデータを取得する
    read_excel_data()
    # SS7データを開き入力CSVファイルを出力する
    export_csv_from_ss7()
    # 入力CSVファイルから既存の特殊荷重入力を含む行を探す
    matching_row, load_count = find_matching_row()
    # 入力CSVファイルにエクセルの荷重表を追加する
    append_load_to_csv(matching_row, load_count)
    # 入力CSVファイルからSS7ファイルを新規作成する
    create_ss7_from_csv()


if __name__ == "__main__":
    main()
